/*
Copyright (C) 2009-2012 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
 */

package org.harctoolbox.IrpMaster;

import com.hifiremote.LibraryLoader;
import com.hifiremote.decodeir.DecodeIRCaller;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.HashSet;

// Note: Arrays represent microsecond data, unless the functions acts on Pronto data.

public class DecodeIR {
    // set debug = 2048 for Debug.debugDecodeIR
    
    private static final String libraryName = "DecodeIR";
    private static boolean libIsLoaded = false;
    
    private static final int invalid = -1;
    
    private static String version;
    
    // Why use com.hifiremote.LibraryLoader at all instead of just loading the library directly?
    // While the constructor of DecodeIRCaller calls com.hifiremote.LibraryLoader which keeps
    // its own HashMap of the libraries it has loaded ITSELF.
    
    
    /**
     * Call without argument to load the shared library either from a architecture dependent
     * subdirectory or from the system's java.library.path.
     * 
     */
    public static void loadLibrary() throws UnsatisfiedLinkError {
        if (libIsLoaded)
            return;

        try {
            // First try the com.hifiremote.LoadLibrary path, i.e. with local,
            // architecture dependent subdirectories...
            LibraryLoader.loadLibrary(null, libraryName);
            libIsLoaded = true;
        } catch (UnsatisfiedLinkError e) {
            Debug.debugDecodeIR("Loading of local " + libraryName
                    + " failed, trying system java.library.path...");
        }

        if (!libIsLoaded) {
            try {
                // ... then try the system path
                LibraryLoader.loadLibrary(libraryName);
                libIsLoaded = true;
                Debug.debugDecodeIR("Loading of " + libraryName + " from system path succeeded.");
            } catch (UnsatisfiedLinkError e) {
                System.err.println("Loading of " + libraryName + " from system path failed.");
                throw e;
            }
        }
        version = (new DecodeIRCaller()).getVersion();
        Debug.debugDecodeIR("LibraryLoader DecodeIR Library path = " + LibraryLoader.getLibraryFolder());
    }
    
    public static class DecodedSignal {
	public String protocolName;
	public int device;
	public int subDevice;
	public int OBC;
	public int[] hex;
	public String miscMessage;
	public String errorMessage;
        private HashMap<String, Integer>miscNames = new HashMap<String, Integer>();

	public DecodedSignal(String protocolName,
			     int device,
			     int subDevice,
			     int OBC,
			     int[] hex,
			     String miscMessage,
			     String errorMessage) {
	    this.protocolName = protocolName;
	    this.device = device;
	    this.subDevice = subDevice;
	    this.OBC = OBC;
	    this.hex = hex;
	    this.miscMessage = miscMessage;
	    this.errorMessage = errorMessage;
            String[] names = miscMessage.replaceAll("\\s*=\\s*", "=").split("[ ,]");
            for (int i = 0; i < names.length; i++) {
                String[] q = names[i].split("=");
                if (q.length == 2) {
                    try {
                        miscNames.put(q[0], Integer.parseInt(q[1]));
                    } catch (NumberFormatException ex) {
                        UserComm.warning("Output from DecodeIR: `" + names[i] + "' could not be parsed.");
                    }
                }
            }
	}

	public DecodedSignal(String protocolName,
			     int device,
			     int subDevice,
			     int OBC) {
	    this(protocolName, device, subDevice, OBC, null, "", "");
	}

	public DecodedSignal(String protocolName,
			     int device,
			     int OBC) {
	    this(protocolName, device, (int) IrpUtils.invalid, OBC);
	}
        
        // There are some idiosyncracises in the output of DecodeIr. I try to take care of them here.
        public boolean isOk(String protocolName, Protocol protocol, HashMap<String, Long> vars) {
            
            if (protocolName == null || protocol == null || vars == null)
                return false;
            
            // First do the protocol specific stuff here ...
            boolean specialOk = true;
            boolean protocolOk = true;
            HashSet<String> exceptionVars = new HashSet();
            if (protocolName.equalsIgnoreCase("x10.n")) {
                protocolOk = this.protocolName.toLowerCase().startsWith("x10:");
                specialOk = Integer.parseInt(this.protocolName.substring(4)) == vars.get("N");
                exceptionVars.add("N");
            } else if (protocolName.endsWith("-???-???")) {
                int length = this.protocolName.lastIndexOf("-") + 1;
                protocolOk = protocolName.startsWith(this.protocolName.toLowerCase().substring(0, length));
                String MNstring = this.protocolName.substring(length);
                String[] s = MNstring.split("\\.");
                specialOk = Integer.parseInt(s[0]) == vars.get("M") && Integer.parseInt(s[1]) == vars.get("N");
                exceptionVars.add("M");
                exceptionVars.add("N");
            } else if (protocolName.equalsIgnoreCase("tivo")) {
                protocolOk = this.protocolName.toLowerCase().startsWith("tivo ");
                specialOk = Integer.parseInt(this.protocolName.substring(10)) == vars.get("U")
                        && this.device == 133
                        && this.subDevice == 48;
                exceptionVars.add("U");
            } else {
                protocolOk = protocolName.equalsIgnoreCase(this.protocolName.replaceAll("\\s", ""));
            }
            
            // ... then the rest is generic
            for (String var : vars.keySet()) {
                
                long deflt = IrpUtils.invalid;
                try {
                    deflt = protocol.getParameterDefault(var, vars);
                } catch (UnassignedException ex) {
                    // This is not an error
                }
                
                if (exceptionVars.contains(var)) {
                    // ignore, due to special properties
                } else if (vars.get(var) == deflt) {
                    // Variable has its default value, nothing should be checked
                } else {
                    // check
                    int decodeIrValue = var.equals("D") ? this.device
                            : var.equals("S") ? this.subDevice
                            : var.endsWith("F") ? this.OBC
                            //: var.equals("I") ? (miscNames.containsKey("ID") ? miscNames.get("ID") : (int) IrpUtils.invalid)
                            : miscNames.containsKey(var) ? miscNames.get(var)
                            : (int) IrpUtils.invalid;
                    if (decodeIrValue == IrpUtils.invalid || vars.get(var).intValue() != decodeIrValue) {
                        // Variable not there or has wrong value, interrupt check as failed
                        return false;
                    }
                }
            }
            return protocolOk && specialOk;
        }

        @Override
	public String toString() {
	    return "protocol = " + protocolName
		+ ", device = " + device
		+ ((subDevice >= 0) ? (", subdevice = " + subDevice) : "")
		+ ", obc = " + OBC
		//+ ", hex = " + getHex
		+ (!miscMessage.equals("") ? (", misc = " + miscMessage) : "")
		+ (!errorMessage.equals("") ? (", error = " + errorMessage) : "");
	}
    }

    private DecodeIRCaller dirc;
    private final static int max_decodes = 10;
    private DecodedSignal[] work;
    private DecodedSignal[] decodedSignals;

    public static String getVersion() {
        loadLibrary();
        return version;
    }
    
    private void setup(int[] us_data, int lengthRepeat, int lengthEnding, int frequency) {
        loadLibrary();
	dirc = new DecodeIRCaller();
	dirc.setBursts(us_data, lengthRepeat, lengthEnding);
	dirc.setFrequency(frequency);
	dirc.initDecoder();
        Debug.debugDecodeIR("DecodeIR was setup with f=" + frequency + ", repeat=" + lengthRepeat
                    + ", ending=" + lengthEnding + ", data=" + IrpUtils.stringArray(us_data));
	
	work = new DecodedSignal[max_decodes];
	int decode_n = 0;
	try {
	    while (dirc.decode()) {
		work[decode_n] = new DecodedSignal(dirc.getProtocolName(),
						     dirc.getDevice(),
						     dirc.getSubDevice(),
						     dirc.getOBC(),
						     dirc.getHex(),
						     dirc.getMiscMessage(),
						     dirc.getErrorMessage());
		decode_n++;
	    }
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.err.println("Only considering the first " + max_decodes + " decodes.");
	}

	decodedSignals = new DecodedSignal[decode_n];
        System.arraycopy(work, 0, decodedSignals, 0, decode_n);
    }

    /**
     * The most general constuctor.
     * @param data Microsecond data, all positive, with odd indexes representing flashes, and even indexes gaps.
     * @param lengthRepeat Length of the repeat sequence, in pairs; thus not always even.
     * @param lengthEnding Length of the ending sequence, in pairs.
     * @param frequency Carrier frequency in Hz.
     */
    public DecodeIR(int[] data, int lengthRepeat, int lengthEnding, int frequency) {
	setup(data, lengthRepeat, lengthEnding, frequency);
    }
    
    /**
     * General constructor.
     * @param irSignal 
     */
    public DecodeIR(IrSignal irSignal) {
	setup(irSignal.toIntArray(), irSignal.getRepeatLength(), irSignal.getEndingLength(),
                (int)Math.round(irSignal.getFrequency()));
    }
    
    /**
     * Constructs a decoder from an array of integers with the Pronto semantics.
     * @param CCF Array of integers with Pronto semantics.
     */
    public DecodeIR(int[] CCF) throws IrpMasterException {
        this(Pronto.ccfSignal(CCF));
    }
 
    /**
     * Constructs a decoder from a Pronto format string.
     * @param ccf Pronto format String.
     */
    public DecodeIR(String ccf) throws IrpMasterException {
	this(Pronto.ccfSignal(ccf));
    }
    
    // Just dispose of signs, better yet would be to check.
    private static int[] parseRawString(String s) throws NumberFormatException {
	String[] work = s.split("\\s+");
	int[] result = new int[work.length];
	for (int i = 0; i < result.length; i++)
            result[i] = Math.abs(Integer.parseInt(work[i].replaceFirst("\\+", "")));

	return result;
    }

    public DecodedSignal[] getDecodedSignals() {
	return decodedSignals;
    }
    
    /**
     * Static version of the constructor with the same arguments. Constructs a DecodeIR object, and applies the getDecodedSignals-function.
     * 
     * @param data
     * @param lengthRepeat
     * @param lengthEnding
     * @param frequency
     * @return array of decodes
     */
    public static DecodedSignal[] decode(int[] data, int lengthRepeat, int lengthEnding, int frequency) {
	return (new DecodeIR(data, lengthRepeat, lengthEnding, frequency)).getDecodedSignals();
    }
    
    /**
     * Static version of the constructor with the same arguments. Constructs a DecodeIR object, and applies the getDecodedSignals-function.
     * @param irSignal
     * @return array of decodes
     */
    public static DecodedSignal[] decode(IrSignal irSignal) {
        return (new DecodeIR(irSignal)).getDecodedSignals();
    }
    
    /**
     * Static version of the constructor with the same arguments. Constructs a DecodeIR object, and applies the getDecodedSignals-function.
     * @param ccf
     * @return array of decodes
     * @throws IrpMasterException 
     */
    public static DecodedSignal[] decode(String ccf) throws IrpMasterException {
        return (new DecodeIR(ccf)).getDecodedSignals();
    }
    
    /**
     * Static version of the constructor with the same arguments. Constructs a DecodeIR object, and applies the getDecodedSignals-function.
     * 
     * @param CCF
     * @return array of decodes
     * @throws IrpMasterException 
     */
    public static DecodedSignal[] decode(int[] CCF) throws IrpMasterException {
        return (new DecodeIR(CCF)).getDecodedSignals();
    }
    
    /**
     * Invokes a DecodeIR object on the argument.
     * @param irSignal 
     */
    public static void invoke(IrSignal irSignal) {
        invoke(irSignal, null, null, null, false);
    }
    
    /**
     * Invokes a DecodeIR object on the first argument, checks if compatible with the data in the rest of the arguments.
     * 
     * @param irSignal
     * @param protocolName
     * @param protocol
     * @param actualParameters
     * @param verbose
     * @return result of the comparison.
     */
    public static boolean invoke(IrSignal irSignal, String protocolName, Protocol protocol, HashMap<String, Long> actualParameters, boolean verbose) {
        if (irSignal == null)
            return false;

        boolean verified = false;
 
        try {
            DecodedSignal[] decodes = decode(irSignal.toIntArray(), irSignal.getRepeatLength(), irSignal.getEndingLength(), (int) irSignal.getFrequency());
            if (decodes == null)
                ;
            else if (decodes.length == 0) {
                if (verbose)
                    System.out.println("-");
            } else
                for (int i = 0; i < decodes.length; i++) {
                    boolean isOk = decodes[i].isOk(protocolName, protocol, actualParameters);
                    if (isOk) {
                        verified = true;
                        Debug.debugDecodeIR(decodes[i] + ": success");
                        // If there is a log file, don't babble.
                        if (verbose)
                            System.out.println("success");
                    } else
                        System.out.println(decodes[i]);
                }
        } catch (UnsatisfiedLinkError ex) {
            UserComm.error("DecodeIR not found in java.library.path");
        }

        return verified;
    }
    
    private static void usage(int exitcode) {
	System.err.println("Usage:");
	System.err.println("DecodeIR [OPTIONS] -v | --version | --help |  CCF-code | raw-code| - | <filename>");
        System.err.println("where OPTIONS = -d|--debug <n>, -i|--intro <n>, -r|--repetition <n>, -e|--ending <n>, -s|--skip <n>, -f|--frequency <frequency in Hz>");
	System.exit(exitcode);
    }

    /**
     * Allows for calling DecodeIR from the command line
     * @param args -- IR signal in Pronto format to be decoded.
     */
    public static void main(String[] args) {
        int debug = 0;
        int frequency = 0;
        int introLength = invalid;
        int repetitionLength = invalid;
        int endingLength = 0;
        try {
            int skip = 0;
            int arg_i = 0;

            BufferedReader reader = null;
            String ccfString = null;
            String rawString = null;

            while (arg_i < args.length && args[arg_i].length() > 1 && args[arg_i].charAt(0) == '-') {
                if (args[arg_i].equals("--help")) {
                    usage(IrpUtils.exitSuccess);
                } else if (args[arg_i].equals("-v") || args[arg_i].equals("--version")) {
                    System.out.println("DecodeIR version " + DecodeIR.getVersion());
                    System.exit(IrpUtils.exitSuccess);
                } else if (args[arg_i].equals("-s") || args[arg_i].equals("--skip")) {
                    arg_i++;
                    skip = Integer.parseInt(args[arg_i++]);
                } else if (args[arg_i].equals("-d") || args[arg_i].equals("--debug")) {
                    arg_i++;
                    debug = Integer.parseInt(args[arg_i++]);
                    Debug.setDebug(debug);
                } else if (args[arg_i].equals("-f") || args[arg_i].equals("--frequency")) {
                    arg_i++;
                    frequency = Integer.parseInt(args[arg_i++]);
                } else if (args[arg_i].equals("-i") || args[arg_i].equals("--intro")) {
                    arg_i++;
                    introLength = Integer.parseInt(args[arg_i++]);
                } else if (args[arg_i].equals("-r") || args[arg_i].equals("--repetition")) {
                    arg_i++;
                    repetitionLength = Integer.parseInt(args[arg_i++]);
                } else if (args[arg_i].equals("-e") || args[arg_i].equals("--ending")) {
                    arg_i++;
                    endingLength = Integer.parseInt(args[arg_i++]);
                } else
                    usage(IrpUtils.exitUsageError);
            }
        
            if (args[arg_i].equals("-"))
                reader = new BufferedReader(new InputStreamReader(System.in));
            else if (args.length - arg_i == 1) {
                if (args[arg_i].startsWith("0000") || args[arg_i].startsWith("0100"))
                    ccfString = args[arg_i];
                else if (args[arg_i].startsWith("+"))
                    rawString = args[arg_i];
                else
                    reader = new BufferedReader(new FileReader(args[arg_i]));
            }

            if (reader != null) {
                for (int i = 0; i < skip; i++)
                    reader.readLine();

                String line = reader.readLine();
                if (line.startsWith("0000") || line.startsWith("0100"))
                    ccfString = line;
                else
                    rawString = line;
            }
            
            DecodedSignal[] result;
            if (ccfString != null) {
                result = decode(ccfString);
            } else if (rawString != null) {
                int[] data = parseRawString(rawString);
                if (introLength != invalid && repetitionLength == invalid)
                    repetitionLength = data.length/2 - introLength - endingLength;
                result = decode(data, repetitionLength, endingLength, frequency);
            } else {
                int[] data = new int[args.length - arg_i];
                if (introLength != invalid && repetitionLength == invalid)
                    repetitionLength = data.length/2 - introLength - endingLength;
                if (args[arg_i].startsWith("+")) {
                    for (int i = 0; i < args.length - arg_i; i++)
                        data[i] = Integer.parseInt(args[i + arg_i].replaceFirst("[\\+-]", ""));
                    result = decode(data, frequency, repetitionLength, endingLength); // ??
                } else {
                    for (int i = 0; i < args.length - arg_i; i++)
                        data[i] = Integer.parseInt(args[i + arg_i], 16);
                    result = decode(data);
                }
            }
            if (result != null)
                for (int i = 0; i < result.length; i++)
                    System.out.println(result[i]);

       	} catch (IrpMasterException ex) {
            System.err.println("IrpMaster exception: " + ex.getMessage());
        } catch (IOException ex) {
            System.err.println(ex.getMessage());
        } catch (NumberFormatException ex) {
            System.err.println("Number format exception " + ex.getMessage());
        } catch (ArrayIndexOutOfBoundsException e) {
	    if (debug > 0)
                System.err.println("ArrayIndexOutOfBoundsException occured");
	    usage(IrpUtils.exitFatalProgramFailure);
	}
    }
}
