/*
Copyright (C) 2011 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
 */
package org.harctoolbox.IrpMaster;

import java.io.*;

/**
 * This class allows for the creation of rendered IR signals in the ICT Format, used by the IRScope.
 *
 */
public class ICT {
    private IrSignal irSignal;
    
    public ICT(IrSignal irSignal) throws IncompatibleArgumentException {
        this.irSignal = irSignal;
    }
    
    private String toPrintString(int[] array) {
        String s = "";
        if (array.length > 0) {
            for (int i = 0; i < array.length; i++) {

                s += (i % 2 == 0)
                        ? ("+" + array[i] + "," + array[i]*irSignal.getFrequency()/1000000.0)
                        : array[i];
                s += "\r\n";
            }
        }
        return s;
    }

    public String toPrintString() {
        String s = "irscope 0\r\ncarrier_frequency " + irSignal.getFrequency()
                + "\r\nsample_count " + (irSignal.getIntroLength() + irSignal.getRepeatLength() + irSignal.getEndingLength()) + "\r\n";
        s += toPrintString(irSignal.getIntroInts()) + toPrintString(irSignal.getRepeatInts()) + toPrintString(irSignal.getEndingInts());
        return s;
    }

    /**
     * Parses an input file into an IRSignal.
     * Note that the ICT file may contain several IR signals in the usual sense.
     * The are returned as one long IRSignal consisting of all the data concatenated.
     *
     * @param reader
     * @return IrSignal consisting of all data in the input concatenated.
     * @throws IOException
     * @throws IncompatibleArgumentException
     */
    public static IrSignal parse(BufferedReader reader) throws IOException, IncompatibleArgumentException {
        int[] data = null;
        String line;
        int index = 0;
        int frequency = 0;
        while ((line = reader.readLine()) != null) {
            String[] args = line.split("[ ,]");
            if (args[0].equals("carrier_frequency"))
                frequency = Integer.parseInt(args[1]);
            else if (args[0].equals("sample_count"))
                data = new int[Integer.parseInt(args[1])];
            else if (args[0].startsWith("+"))
                data[index++] = Integer.parseInt(args[0].substring(1));
            else if (args[0].startsWith("-"))
                //data[index++] = (int) (((double) Integer.parseInt(args[0].substring(1)) * frequency) / 1000000.0 + 0.5);
                data[index++] = Integer.parseInt(args[0].substring(1));
        }
        return new IrSignal((double) frequency, (double) IrpUtils.invalid, new IrSequence(data), null, null);
    }

    public static IrSignal parse(String filename) throws FileNotFoundException, IOException, IncompatibleArgumentException {
        return parse(new BufferedReader(new InputStreamReader(new FileInputStream(filename))));
    }
    
    public static IrSignal parse(File file) throws FileNotFoundException, IOException, IncompatibleArgumentException {
        return parse(new BufferedReader(new InputStreamReader(new FileInputStream(file))));
    }

    public static void main(String[] args) {
        try {
            IrSignal ip = parse(args[0]);
            System.out.println(ip.ccfString());
        } catch (IncompatibleArgumentException ex) {
            System.err.println(ex.getMessage());
        } catch (FileNotFoundException ex) {
            System.err.println(ex.getMessage());
        } catch (IOException ex) {
            System.err.println(ex.getMessage());
        }
    }
}
