/*
Copyright (C) 2011 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
 */
package org.harctoolbox.IrpMaster;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.LinkedHashMap;

/**
 * This class creates configuration files for LIRC.
 *
 * @author Bengt Martensson
 */
public class LircExport {

    private String name;
    private String comment;
    private int frequency;
    private int gap;
    private LinkedHashMap<String, int[]> signals;

    /**
     * Constructor for the class.
     *
     * @param name Name of the "remote", as in the name file in the LIRC remote
     * @param comment String Comment to go into the generated file.
     * @param frequency Modulation frequency in Hz.
     */
    public LircExport(String name, String comment, double frequency) {
        this.name = name;
        this.comment = comment;
        this.frequency = (int) frequency;
        gap = 0;
        signals = new LinkedHashMap<String, int[]>();
    }

    /**
     * Adds a signal to the export, to be a command in the remote.
     *
     * @param name Name of the command
     * @param data
     */
    public void addSignal(String name, int[] data) {
        signals.put(name, data);
        gap = Math.max(gap, data[data.length-1]);
    }

    /**
     * Adds a signal to the export, to be a command in the remote.
     *
     * @param parameters Parameters, to form a name of the LIRC command
     * @param irSignal IrSignal to be the data of the command.
     */
    public void addSignal(HashMap parameters, IrSignal irSignal) {
        String[] keys = (String[]) parameters.keySet().toArray(new String[0]);
        java.util.Arrays.sort(keys, String.CASE_INSENSITIVE_ORDER);
        String sigName = "";
        for (int i = 0; i < keys.length; i++)
            sigName += keys[i] + String.format("%03d", parameters.get(keys[i]));
        addSignal(sigName.toLowerCase(), irSignal.toIntArray());
    }

    /**
     * Write the instance to a the file in the argument.
     * @param file File to be written.
     * @return Success status.
     */
    public boolean write(File file) {
        PrintStream stream = null;
        try {
            stream = new PrintStream(file);
        } catch (FileNotFoundException ex) {
            System.err.println("Cannot open Lirc export file " + file);
            return false;
        }
        write(stream);
        stream.close();
        return true;
    }

    /**
     * Write the instance to a the PrintStream in the argument.
     * 
     * @param stream PrintStream, should be ready for writing.
     */
    public void write(PrintStream stream) {
        stream.println("#");
        stream.println("# " + comment);
        stream.println("#");
        stream.println("begin remote");
        stream.println("\tname\t" + name);
        stream.println("\tflags\tRAW_CODES");
        stream.println("\teps\t30");
        stream.println("\taeps\t100");
        stream.println("\tgap\t" + gap);
        stream.println("\t\tbegin raw_codes");
        for (String sigName : signals.keySet()) {
            stream.println();
            stream.print("\t\t\tname " + sigName);
            int[] signal = signals.get(sigName);
            for (int i = 0; i < signal.length - 1; i++) {
                if (i % 8 == 0) {
                    stream.println();
                    stream.print("\t\t\t\t");
                }
                stream.print(signal[i] + " ");
            }
        }
        stream.println();
        stream.println("\t\tend raw_codes");
        stream.println("end remote");
    }
}
