/*
Copyright (C) 2012 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
 */

package org.harctoolbox.IrpMaster;

import java.io.*;
import java.util.ArrayList;
import org.antlr.runtime.RecognitionException;

/**
 * This class models a macro, consisting of a sequence of IrSignals,
 * each with a number of repetitions, possibly divided by delays.
 * There is only one output function, taking the macro to a ModulatedIrSequence.
 * 
 * <p>Possible future extension: Allow something like beginrepeat, beginending that will instead
 * output an IrSignal, allowing arbitrary repetitions.
 */
public class Macro {
    private ArrayList<ModulatedIrSequence>sequences = new ArrayList<ModulatedIrSequence>();

    /**
     * Constructs a Macro from the arguments.
     * @param irProtocolsIniPath
     * @param reader
     * @throws IrpMasterException
     * @throws FileNotFoundException
     * @throws RecognitionException
     * @throws IOException
     * @throws UnassignedException
     */
    public Macro(String irProtocolsIniPath, BufferedReader reader) throws IrpMasterException, FileNotFoundException, RecognitionException, IOException, UnassignedException {
        //IrSequence result = new IrSequence();
        String line;
        //ModulatedIrSequence last = null;
        //int lastIndex = (int) IrpUtils.invalid;
        while ((line = reader.readLine()) != null) {
            String[] args = line.split("\\s+");
            int arg_i = 0;
            int repetitions = 0;
            if (args[arg_i].equals("-r")) {
                arg_i++;
                repetitions = Integer.parseInt(args[arg_i++]);
            }
            if (args.length - arg_i == 0) {
                // empty line, just proceed to the next line
            } else if (args.length - arg_i == 1) {
                // just one number, consider it a pure delay, in milliseconds
                double delay = 1000*Double.parseDouble(args[arg_i]);
                if (sequences.size() > 0) {
                    int lastIndex = sequences.size() - 1;
                    sequences.set(lastIndex, sequences.get(lastIndex).append(delay));
                }
            } else {
                // none of the above, parse it according to the smart parsing
                IrSignal irSignal = new IrSignal(irProtocolsIniPath, arg_i, args);
                ModulatedIrSequence irSequence = irSignal.toModulatedIrSequence(repetitions);
                sequences.add(null);
            }
        }
    }

    /**
     * Constructs a Macro from the arguments.
     * @param irProtocolsIniPath
     * @param macro
     * @throws IrpMasterException
     * @throws FileNotFoundException
     * @throws IOException
     * @throws RecognitionException
     * @throws UnassignedException
     */
    public Macro(String irProtocolsIniPath, String macro) throws IrpMasterException, FileNotFoundException, IOException, RecognitionException, UnassignedException {
        this(irProtocolsIniPath, new BufferedReader(new StringReader(macro)));
    }

    /**
     * Constructs a Macro from the arguments.
     * @param irProtocolsIniPath
     * @param file
     * @throws FileNotFoundException
     * @throws IrpMasterException
     * @throws RecognitionException
     * @throws IOException
     * @throws UnassignedException
     */
    public Macro(String irProtocolsIniPath, File file) throws FileNotFoundException, IrpMasterException, RecognitionException, IOException, UnassignedException {
        this(irProtocolsIniPath, new BufferedReader(new FileReader(file)));
    }

    /**
     * 
     * @return ModulatedIrSequence representing the macro.
     */
    public ModulatedIrSequence toModulatedIrSequence() {
        return new ModulatedIrSequence(this.sequences.toArray(new ModulatedIrSequence[0]));
    }
}
