/*
Copyright (C) 2012 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
*/

package org.harctoolbox.IrpMaster;

/**
 * A ModulatedIrSequence is an IrSequence with the additional properties of a modulation frequency and a duty cycle.
 * The name is slightly misleading since the modulation frequency can be 0, it just to be present.
 */
public class ModulatedIrSequence extends IrSequence {

    /** Modulation frequency in Hz */
    protected double frequency = 0;

    /** Duty cycle of the modulation, a number between 0 and 1. Use -1 for unassigned.*/
    protected double dutyCycle = (double) IrpUtils.invalid;

    /**
     *
     * @return modulation frequency in Hz.
     */
    public double getFrequency() {
        return frequency;
    }

    /**
     *
     * @return Duty cycle.
     */
    public double getDutyCycle() {
        return dutyCycle;
    }

    public ModulatedIrSequence() {
    }

    /**
     * Constructs a ModulatedIrSequence from its arguments.
     *
     * @param irSequence irSequence to be copied from
     * @param frequency
     * @param dutyCycle
     */
    public ModulatedIrSequence(IrSequence irSequence, double frequency, double dutyCycle) {
        data = irSequence.data;
        this.frequency = frequency;
        this.dutyCycle = dutyCycle;
    }

    /**
     * Concatenates the IrSequences in the argument to a new sequence.
     * Frequency and duty cycle are set to the average between minimun and maximum values by th components, if it makes sense.
     * @param seqs One or more ModulatedIrSequences
     */
    public ModulatedIrSequence(ModulatedIrSequence... seqs) {
        int cumulatedLength = 0;
        double minf = Double.MAX_VALUE;
        double maxf = Double.MIN_VALUE;
        double mindc = Double.MAX_VALUE;
        double maxdc = Double.MIN_VALUE;
        for (ModulatedIrSequence seq : seqs) {
            minf = Math.min(minf, seq.frequency);
            maxf = Math.max(maxf, seq.frequency);
            mindc = Math.min(mindc, seq.frequency);
            maxdc = Math.max(maxdc, seq.frequency);
            cumulatedLength += seq.getLength();
        }
        
        dutyCycle = mindc > 0 ? (mindc + maxdc)/2 : (double) IrpUtils.invalid;
        frequency = minf > 0 ? (minf + maxf)/2 : 0;
        data = new double[cumulatedLength];
        int beginIndex = 0;
        for (ModulatedIrSequence seq : seqs) {
            System.arraycopy(seq.data, 0, data, beginIndex, seq.data.length);
            beginIndex += seq.data.length;
        }
    }

    public IrSignal toIrSignal() {
        return new IrSignal(frequency, dutyCycle, this, new IrSequence(), new IrSequence());
    }
    
    /**
     * Appends a delay to the end of the ModulatedIrSequence. Original is left untouched.
     * @param delay microseconds of silence to be appended to the IrSequence.
     * @return Copy of object with additional delay at end.
     */
    @Override
    public ModulatedIrSequence append(double delay) throws IncompatibleArgumentException {
        IrSequence irSequence = ((IrSequence) this).append(delay);
        return new ModulatedIrSequence(irSequence, frequency, dutyCycle);
    }
}
